# -*- coding: utf-8 -*-

"""
BLIP ELAN Language Annotation (BELA) package
"""

# This code is a part of BELA package: https://github.com/letuananh/bela
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from .__version__ import __author__, __email__, __copyright__, __maintainer__
from .__version__ import __credits__, __license__, __description__, __url__
from .__version__ import __version_major__, __version_long__, __version__, __status__

from .common import maketime, getlang
from .common import is_special_token
from .common import process_text, tokenize
from .common import KNOWN_LANGUAGES, KNOWN_LANGUAGE_CLASSES
from .bela1 import Bela1
from .bela2 import Bela2, SPECIAL_SPEAKER, SPECIAL_SPEAKER_NAME
from .bela2 import read_eaf, from_elan
from .builder import Builder

__all__ = ['Bela1', 'maketime', 'getlang', 'is_special_token',
           'process_text', 'tokenize',
           'KNOWN_LANGUAGES', 'KNOWN_LANGUAGE_CLASSES',
           'SPECIAL_SPEAKER', 'SPECIAL_SPEAKER_NAME',
           'Bela2', 'read_eaf', 'from_elan',
           'Builder',
           "__version__", "__author__", "__description__", "__copyright__"]



# -*- coding: utf-8 -*-

# This code is a part of BELA package: https://github.com/letuananh/bela
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from pathlib import Path
from speach import elan
from .bela2 import Bela2

"""
BELA transcription builder
"""

_TEMPLATE_PATH = Path(__file__).parent / 'data' / 'bela_v3_blank.eaf'


def _create_blank_bela():
    return elan.read_eaf(_TEMPLATE_PATH)


class Builder:
    """ BELA transcript builder
    """

    def __init__(self, add_marker=False, *args, **kwargs):
        """ Return a blank BELA2 transcript builder object
        """
        self.eaf = _create_blank_bela()
        if add_marker:
            self.add_marker()

    def add_marker(self):
        self.eaf.new_tier('ActivityMarkers', 'Activity_Marker')

    def create_participant(self, name, participant='', target=True, matrix=True, sensitive=True):
        ''' Create tier structure for a new participant '''
        self.eaf.new_tier(f'{name} (Utterance)', 'Utterance', participant=participant)
        self.eaf.new_tier(f'{name} (Chunk)', 'Chunk_TimeSub', f'{name} (Utterance)', participant=participant)
        self.eaf.new_tier(f'{name} (Language)', 'Lang_SymAssoc', f'{name} (Chunk)', participant=participant)
        if target:
            self.eaf.new_tier(f'{name} (Target_EL)', 'Target_Words_Eng_IncIn', f'{name} (Chunk)', participant=participant)
            self.eaf.new_tier(f'{name} (Target_CL)', 'Target_Words_Mandarin_IncIn', f'{name} (Chunk)', participant=participant)
            self.eaf.new_tier(f'{name} (Target_ML)', 'Target_Words_Malay_IncIn', f'{name} (Chunk)', participant=participant)
            self.eaf.new_tier(f'{name} (Target_TL)', 'Baby_Target_Words_IncIn', f'{name} (Chunk)', participant=participant)
        self.eaf.new_tier(f'{name} (Translation)', 'Translation', f'{name} (Utterance)', participant=participant)
        if matrix:
            self.eaf.new_tier(f'{name} (Matrix)', 'Matrix_Lang', f'{name} (Utterance)', participant=participant)
        if sensitive:
            self.eaf.new_tier(f'{name} (Sensitive_Masking)', 'Sensitive_Masking_IncIn', f'{name} (Utterance)', participant=participant)

    def annotate(self, name, text, from_ts, to_ts, chunks, times, trans):
        ''' Create a new annotation '''
        pu = self.eaf[f'{name} (Utterance)']
        pc = self.eaf[f'{name} (Chunk)']
        pl = self.eaf[f'{name} (Language)']
        pt = self.eaf[f'{name} (Translation)']
        u = pu.new_annotation(text, from_ts, to_ts)
        pt.new_annotation(trans, ann_ref_id=u.ID)
        chunk_texts = [c[0] for c in chunks]
        chunk_langs = [c[1] for c in chunks]
        chunk_objs = pc.new_annotation(None, values=chunk_texts, timeslots=times, ann_ref_id=u.ID)
        for c, l in zip(chunk_objs, chunk_langs):
            pl.new_annotation(l, ann_ref_id=c.ID)
        return pu

    def to_csv_rows(self, *args, **kwargs):
        """ Generate CSV content """
        return self.eaf.to_csv_rows(*args, **kwargs)

    def to_xml_bin(self, *args, **kwargs):
        """ Generate EAF content (bytes) in XML format """
        return self.eaf.to_xml_bin(*args, **kwargs)

    def to_xml_str(self, *args, **kwargs):
        """ Generate EAF content string in XML format """
        return self.eaf.to_xml_str(*args, **kwargs)

    def save(self, *args, **kwargs):
        ''' Write BELA transcript to an EAF file '''
        return self.eaf.save(*args, **kwargs)

    def to_bela(self, *args, **kwargs):
        return Bela2.from_elan(self.eaf.clone(), *args, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Test script for BELA
Latest version can be found at https://github.com/letuananh/bela

:copyright: (c) 2020 Le Tuan Anh <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import os
import unittest
import logging

import bela

TEST_DIR = os.path.dirname(os.path.realpath(__file__))


def getLogger():
    return logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# Test cases
# ------------------------------------------------------------------------------

class TestBELA(unittest.TestCase):

    def test_bela(self):
        print(f"Testing BELA {bela.__version__}")
        self.assertTrue(bela.__version__)



# -*- coding: utf-8 -*-
# BELA's package version information

# This code is a part of BELA package: https://github.com/letuananh/bela
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

__author__ = "Le Tuan Anh"
__email__ = "tuananh.ke@gmail.com"
__copyright__ = "Copyright (c) 2019, Le Tuan Anh, BLIP lab NTU, Singapore"
__credits__ = []
__license__ = "MIT License"
__description__ = "BLIP ELAN Language Annotation package"
__url__ = "https://bela.readthedocs.io/"
__maintainer__ = "Le Tuan Anh"
__version_major__ = "2.0.0"
__version__ = "{}a21.post6".format(__version_major__)
__version_long__ = "{} - Alpha 21.post6".format(__version_major__)
__status__ = "3 - Alpha"



# -*- coding: utf-8 -*-

# This code is a part of BELA package: https://github.com/letuananh/bela
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

"""
BELA version 2.x convention
"""

import logging
import re
import types

from collections import deque
from chirptext import DataObject
from speach import elan

from .common import TIER_PARSER
from .common import tokenize
from .common import LanguageMix
from .common import _find_invalid_characters


def getLogger():
    return logging.getLogger(__name__)


SPECIAL_TIERS = ['ActivityMarkers']
SPECIAL_SPEAKER_NAME = 'Transcriber'
SPECIAL_SPEAKER = ':transcriber:'
DEFAULT_TURN_THRESHOLD = 1500
PTN_ANSI_CHUNK = re.compile(r"^\s*(([A-Za-z\'\:\?#\=_\.\-~\+@]+|\:v\:l\:iso\d{3}_\d_[a-z]{3})\s*)+\s*$")
ANSI_CHECK_LANGUAGES = ('English', 'Vocal Sounds', 'Malay', 'Red Dot', ':v:airstream', ':v:crying', ':v:vocalizations')


def parse_tier_name(name):
    ''' Parse BELA tier name convention '''
    m = TIER_PARSER.match(name)
    if m:
        return m.group('person'), m.group('tier')
    else:
        return None, None


def time_map(tier):
    _time_map = {}
    for u in tier:
        _time_map[(u.from_ts, u.to_ts)] = u
        if u.from_ts is not None and u.from_ts.value is not None \
           and u.to_ts is not None and u.to_ts.value is not None:
            _time_map[(u.from_ts.value, u.to_ts.value)] = u
    return _time_map


def _enforce_ansi_alphabets(cu, languages=ANSI_CHECK_LANGUAGES):
    ''' [Internal] Ensure that chunks of certain languages (e.g. English, Malay, etc.) 
    only contains English alphabets & BELA's special characters

    End users should NOT use this function.
    It is not a part of BELA's standard APIs and may be removed in the future.
    '''
    if cu.language in languages:
        if cu.text:
            parts = cu.text.split()
            for p in parts:
                m = PTN_ANSI_CHUNK.match(p)
                if m is None:
                    cu.warnings.append(f"Unusual character was found in chunk '{cu.text}' ('{cu.language}')")
                    return


def _validate_baby_language(cu, person_name='unknown_participant'):
    ''' [Internal] Validate baby language annotations
    Show warnings when babies' chunks are tagged with adult languages

    End users should NOT use this function.
    It is not a part of BELA's standard APIs and may be removed in the future.
    '''
    if cu.text.strip() == '###':
        _name = person_name.strip()
        _lang = cu.language.strip()
        if _name.startswith('Baby'):
            if _lang not in (':v:vocalizations', ':v:airstream', ':v:crying', ':v:laughter'):
                cu.errors.append(f"'###' is tagged as '{cu.language}' for {_name} (should be in (':v:vocalizations', ':v:airstream', ':v:crying', ':v:laughter'))")
        elif _name.startswith('Sibling'):
            if _lang not in (':v:vocalizations', ':v:airstream', ':v:crying', ':v:laughter'):
                if cu.warnings is None:
                    cu.warnings = []
                cu.warnings.append(f"'###' is tagged as '{cu.language}' for {_name} (should be in (':v:vocalizations', ':v:airstream', ':v:crying', ':v:laughter'))")
        elif _lang != 'Vocal Sounds':
            cu.errors.append(f"'###' is tagged as '{cu.language}' for {_name} (should be 'Vocal Sounds' for non-Baby participants)")


def _utterance_tokenize(self, *args, **kwargs):
    """ [Internal] Tokenize an utterance or a chunk

    End users should NOT use this function.
    It is not a part of BELA's standard APIs and may be removed in the future.
    """
    return tokenize(self.value, *args, **kwargs)


def _map_children(parent_tier, child_tier, errors=None, tier_class='chunks', is_many=True):
    ''' map all children chunks into parent's utterances '''
    _parents = sorted(parent_tier, key=lambda x: x.from_ts)
    _children = deque(sorted(child_tier, key=lambda x: x.from_ts))
    for ann in _parents:
        while _children:
            ann_child = _children[0]
            if ann_child.from_ts >= ann.from_ts and ann_child.to_ts <= ann.to_ts:
                if is_many:
                    if not getattr(ann, tier_class):
                        setattr(ann, tier_class, [])
                    getattr(ann, tier_class).append(ann_child)
                else:
                    if getattr(ann, tier_class) is None:
                        setattr(ann, tier_class, ann_child)
                    else:
                        getLogger().warning("Conflicting child annotations were found")
                        if errors is not None:
                            errors.append("Conflicting child annotations were found")
                _children.popleft()
                getLogger().debug(f"linked: {repr(ann)} -- {repr(ann_child)}")
            elif ann.from_ts > ann_child.to_ts:
                getLogger().warning(f"DISCARDED -- {repr(ann_child)}")
                if errors is not None:
                    errors.append(f"Orphaned annotation found -- (#{ann_child.ID}) {repr(ann_child)}")
                _children.popleft()
            else:
                getLogger().debug(f"no relation: {repr(ann)} -- {repr(ann_child)}")
                break
    if _children:
        for child in _children:
            getLogger().debug(f"Orphaned annotation found: {child}")


class Person(DataObject):
    def __init__(self, name, code=None, utterances=None, tiers=[], belav2=None, **kwargs):
        super().__init__(name=name, code=code, utterances=utterances, **kwargs)
        self.belav2 = belav2
        if utterances is None:
            self.utterances = elan.Tier()
        self.__tier_map = {}
        self.__tiers = list(tiers) if tiers else []

    @property
    def tiers(self):
        return self.__tiers

    @property
    def tier_classes(self):
        return list(self.__tier_map.keys())

    def __contains__(self, key):
        return key in self.__tier_map

    def __getitem__(self, key):
        return self.__tier_map[key] if key in self.__tier_map else None

    def __iter__(self):
        return iter(self.tiers)

    def add_tier(self, tier):
        if tier in self.tiers:
            return
        if not tier.tier_class:
            if self.belav2 is not None:
                self.belav2.errors.append(f"Tier [{tier.ID}] does not have a tier_class")
            else:
                getLogger().warning(f"Tier [{tier.ID}] does not have a tier_class")
        elif tier.tier_class in self.__tier_map:
            if self.belav2 is not None:
                self.belav2.errors.append(f"User [{self.code}] has more than one [class={tier.tier_class}] tier")
            else:
                getLogger().warning(f"User [{self.code}] has more than one [class={tier.tier_class}] tier")
        else:
            self.__tier_map[tier.tier_class] = tier
        self.tiers.append(tier)

    def __str__(self):
        return f"Person(name={repr(self.name)}, code={repr(self.code)})"


class Bela2(DataObject):
    ''' BELA-convention version 2
    '''

    def __init__(self, elan, path=":memory:", allow_empty=False,
                 nlp_tokenizer=False, word_only=True, ellipsis=True,
                 validate_baby_languages=False,
                 ansi_languages=ANSI_CHECK_LANGUAGES,
                 auto_tokenize=True,
                 split_punc=True, remove_punc=True, **kwargs):
        ''' Create a new Bela2 object from an :class:`speach.elan.ELANDoc` object

        :param: elan: An ELANDoc object
        :type: elan: speach.elan.ELANDoc

        :returns: a Bela2 object
        :rtype: bela.Bela2
        '''
        super().__init__(elan=elan, path=path, errors=[], warnings=[],
                         validate_baby_languages=validate_baby_languages,
                         ansi_languages=ansi_languages,
                         auto_tokenize=auto_tokenize,
                         allow_empty=allow_empty, **kwargs)
        self.__person_map = {}
        # create special speaker map (i.e. Transcriber)
        self.__person_map[SPECIAL_SPEAKER] = Person(name=SPECIAL_SPEAKER_NAME, code=SPECIAL_SPEAKER, belav2=self)
        self.__persons = None
        self.__participant_codes = None
        self.__word_only = word_only
        self.__nlp_tokenizer = nlp_tokenizer
        self.__ellipsis = ellipsis
        self.__split_punc = split_punc
        self.__remove_punc = remove_punc
        if elan is not None:
            self.parse_names()
            self._init_tier_map()
            if self.auto_tokenize:
                self.tokenize()

    def tiers(self):
        return self.elan.tiers()

    @property
    def roots(self):
        ''' Direct access to all underlying ELAN root tiers '''
        return self.elan.roots

    @property
    def media_file(self):
        return self.elan.media_file

    @property
    def media_url(self):
        return self.elan.media_url

    @property
    def relative_media_url(self):
        return self.elan.relative_media_url

    @property
    def annotation(self, annID):
        ''' Get an annotation object by ID '''
        return self.elan.annotation(annID)

    @property
    def person_map(self):
        ''' Map participant (i.e. person code) to person object '''
        return self.__person_map

    @property
    def persons(self):
        ''' All Person objects in this BELA object '''
        if self.__persons is None:
            self.__persons = tuple(self.__person_map.values())
        return self.__persons

    @property
    def participant_codes(self):
        ''' Immutable list of participant codes '''
        if self.__participant_codes is None:
            # TODO: Make this thread safe?
            self.__participant_codes = tuple(i for i in self.person_map.keys())
        return self.__participant_code

    @property
    def word_only(self):
        return self.__word_only

    def count_sents(self):
        utt_count = 0
        for tier in self.elan:
            if tier.tier_class == "Utterance":
                utt_count += len(tier)
        return utt_count

    def count_chunks(self):
        chunk_count = 0
        for tier in self.elan:
            if tier.tier_class == "Chunk":
                chunk_count += len(tier)
        return chunk_count

    def get_language_set(self):
        languages = set()
        for tier in self.elan:
            if tier.tier_class == "Language":
                for ann in tier:
                    languages.add(ann.value)
        return set(languages)

    def parse_name(self, tier):
        ''' (Internal) Parse participant name and tier type from a tier object and then update the tier object

        This function is internal and should not be used outside of this class.

        :param tier: The tier object to parse
        :type tier: speach.elan.ELANTier
        '''

        if tier.ID in SPECIAL_TIERS:
            tier.tier_class = tier.ID
            tier.speaker_name = SPECIAL_SPEAKER
        else:
            speaker_name, tier_class = parse_tier_name(tier.ID)
            if speaker_name and tier_class:
                tier.tier_class = tier_class
                tier.speaker_name = speaker_name
            else:
                self.errors.append(f"Invalid tier name: {tier.ID}")

    def _init_tier_map(self):
        ''' Construct BELA info structure from EAF '''
        # init roots first
        for tier in self.roots:
            if tier.tier_class == 'Utterance':
                if not tier.participant:
                    self.errors.append(f"Tier [{tier.ID}] does not have participant code")
                elif tier.participant in self.__person_map:
                    self.errors.append(f"Person [{tier.participant}] has more than one utterance tier")
                else:
                    person = Person(tier.speaker_name, code=tier.participant, utterances=tier, belav2=self)
                    self.__person_map[tier.participant] = person
            elif tier.ID not in SPECIAL_TIERS:
                self.errors.append(f"Unknown root tier: {tier.ID}")
        # init other tiers
        for tier in self.elan:
            # verify timestamps
            if tier.time_alignable:
                for ann in tier:
                    if ann.from_ts is None or ann.from_ts.value is None or ann.to_ts is None or ann.to_ts.value is None:
                        self.errors.append(f"Annotation with corrupted timestamp: {ann.value} (Timestamp: {ann.from_ts} -- {ann.to_ts}) | Tier: {tier.ID}")
                        if ann.errors is None:
                            ann.errors = []
                        ann.errors.append(f"Corrupted timestamp: {ann.value} | (Timestamp: {ann.from_ts} -- {ann.to_ts})")
                        if ann.from_ts is not None or (ann.from_ts.value is None and ann.to_ts is not None):
                            ann.from_ts.value = ann.to_ts.value
                            ann.errors.append("Assumed from_ts value from to_ts")
                        if ann.to_ts is not None and ann.to_ts.value is None and ann.from_ts is not None:
                            ann.to_ts.value = ann.from_ts.value
                            ann.errors.append("Assumed to_ts value from from_ts")
            if tier.ID in SPECIAL_TIERS:
                self.__person_map[SPECIAL_SPEAKER].tiers.append(tier)
            elif tier.participant not in self.__person_map:
                self.errors.append(f"Unknown person code [{tier.participant}] used in tier [{tier.ID}]")
            else:
                self.__person_map[tier.participant].add_tier(tier)
        # link languages if available
        for person in self.persons:
            if person.utterances:
                for u in person.utterances:
                    u.person = person
                if person['Chunk']:
                    _map_children(person.utterances, person['Chunk'], errors=self.errors, tier_class='chunks')
                else:
                    self.errors.append(f"Person {person.name} ({person.code}) does not have a chunk tier")
                if not person['Language']:
                    self.errors.append(f"Person {person.name} ({person.code}) does not have a language tier")
                if 'Translation' in person:
                    translations = person['Translation']
                    _translation_map = {u.ref: u.value for u in translations}
                    for u in person.utterances:
                        if u in _translation_map:
                            if not u.translation:
                                u.translation = _translation_map[u]
                            else:
                                self.errors.append(f"Conflicted translation for [{person}] Time: [{u.from_ts} -- {u.to_ts}]")
                else:
                    self.errors.append(f"Person {person.name} ({person.code}) does not have a translation tier")
            if person['Chunk'] and person['Language']:
                lang_tier_time_map = time_map(person['Language'])
                linked_language_annotations = set(person['Language'])
                for cu in person['Chunk']:
                    key = (cu.from_ts, cu.to_ts)
                    if key in lang_tier_time_map:
                        cu.language = lang_tier_time_map[key].value
                        linked_language_annotations.remove(lang_tier_time_map[key])
                    elif cu.from_ts is not None and cu.from_ts.value is not None \
                        and cu.to_ts is not None and cu.to_ts.value is not None:
                        # [2021-09-07 火 14:32]
                        # [TA] try to map using timestamp values if possible
                        key = (cu.from_ts.value, cu.to_ts.value)
                        if key in lang_tier_time_map:
                            cu.language = lang_tier_time_map[key].value
                            linked_language_annotations.remove(lang_tier_time_map[key])
                if linked_language_annotations:
                    for ann in linked_language_annotations:
                        self.errors.append(f"Orphaned language annotation could not be linked: {ann} [{ann.from_ts} -- {ann.to_ts}]")
            # validate text from utterance tier and chunk tier
            if person.utterances:
                for u in person.utterances:
                    # bind tokenize() function to utterances
                    u.tokenize = types.MethodType(_utterance_tokenize, u)
                    if u.errors is None:
                        u.errors = []
                    if u.warnings is None:
                        u.warnings = []
                    if not u.text.strip():
                        if not self.allow_empty:
                            u.errors.append(f"Empty annotation '' found at [{u.from_ts} :: {u.to_ts}]")
                        else:
                            u.warnings.append(f"Empty annotation '' found at [{u.from_ts} :: {u.to_ts}]")
                    if u.chunks:
                        for cu in u.chunks:
                            # bind tokenize() function to chunks
                            cu.tokenize = types.MethodType(_utterance_tokenize, cu)
                            if cu.errors is None:
                                cu.errors = []
                            if cu.warnings is None:
                                cu.warnings = []
                            if self.ansi_languages:
                                _enforce_ansi_alphabets(cu, languages=self.ansi_languages)
                            if self.validate_baby_languages:
                                _validate_baby_language(cu, person.name)
                            if not cu.text.strip():
                                if u.text or not self.allow_empty:
                                    u.errors.append(f"Empty chunk annotation '' found at [{cu.from_ts} :: {cu.to_ts}]")
                                else:
                                    u.warnings.append(f"Empty chunk annotation '' found at [{cu.from_ts} :: {cu.to_ts}]")
                            if cu.language is None or not cu.language.strip():
                                if not self.allow_empty or cu.text.strip() or u.text.strip():
                                    u.errors.append(f"Language tag not found in the chunk `{cu.text.strip()}` [{cu.from_ts} :: {cu.to_ts}]")
                                else:
                                    u.warnings.append(f"Language tag not found in the chunk `{cu.text.strip()}` [{cu.from_ts} :: {cu.to_ts}]")
                            elif "#!" in cu.language:
                                u.errors.append(f"Unsure language tag ({cu.language}) was used for chunk `{cu.text.strip()}` [{cu.from_ts} :: {cu.to_ts}]")
                    else:
                        u.chunks = []  # [2022-03-16 水 11:38][TA] Make sure that chunks is not None
                    u_value = u.text.replace(' ', '')
                    _chunks = u.chunks if u.chunks else []
                    c_value = ''.join(x.text for x in _chunks)
                    c_value = c_value.replace(' ', '')
                    if u_value != c_value:
                        _chunk_texts = ' '.join(x.text.strip() for x in _chunks)
                        # logging.getLogger(__name__).info(f"mismatch:\n   + u_value: {repr(u_value)}\n   + c_value: {repr(c_value)}")
                        u.errors.append(f"Utterance text and chunks are mismatched ({repr(u.text)} != {repr(_chunk_texts)})")

    def parse_names(self):
        for tier in self.elan:
            self.parse_name(tier)

    def tokenize(self):
        ''' tokenize all utterances '''
        for tier in self.elan:
            if tier.tier_class in ('Utterance', 'Chunk'):
                for ann in tier:
                    ann_errors = [] if ann.errors is None else ann.errors
                    ann.words = tokenize(ann.value, language=ann.language,
                                         errors=ann_errors,
                                         ellipsis=self.__ellipsis,
                                         nlp_tokenizer=self.__nlp_tokenizer,
                                         split_punc=self.__split_punc,
                                         remove_punc=self.__remove_punc,
                                         word_only=self.__word_only)
                    _invalid_chars = _find_invalid_characters(ann.value, language=ann.language)
                    if _invalid_chars:
                        if ann.language:
                            ann_errors.append(f"Invalid characters, new line, or tab found ({repr(_invalid_chars)}) (language: {ann.language})")
                        else:
                            ann_errors.append(f"Invalid characters, new line, or tab found ({repr(_invalid_chars)})")
                    ann.errors = ann_errors
                pass

    def find_turns(self, threshold=DEFAULT_TURN_THRESHOLD):
        ''' Find potential turn-takings

        :param threshold: Delay between utterances in milliseconds
        :type threshold: float

        :return: List of utterance pairs (2-tuple) (from utterance, to utterance object)
        '''
        _utterances = []
        for person_code, person in self.person_map.items():
            if person.utterances:
                for u in person.utterances:
                    u.person = person
                    _utterances.append(u)
        _utterances.sort(key=lambda x: x.from_ts)
        # is_turn = False
        _turns = []
        for idx, u in enumerate(_utterances):
            if idx == len(_utterances) - 1:
                continue
            next_u = _utterances[idx + 1]
            if u.to_ts is None or u.to_ts.value is None or \
               next_u.from_ts is None or next_u.from_ts.value is None:
                continue
            delta_t = u.to_ts - next_u.from_ts
            if next_u.person != u.person and abs(delta_t) <= threshold:
                _turns.append((u, next_u))
                # is_turn = True
        return _turns

    def to_language_mix(self, to_ts=None, auto_compute=True):
        ''' Collapse utterances to generate a language mix timeline '''
        langmix = LanguageMix()
        for person in self.persons:
            if not person.utterances:
                continue
            getLogger().debug(f"{person.name}  -- {person.code}")
            for u in person.utterances:
                if u.chunks:
                    for c in u.chunks:
                        if c.from_ts and c.to_ts:
                            if to_ts is not None and c.to_ts > to_ts:
                                continue
                            else:
                                langmix.add(c)
        return langmix.compute() if auto_compute else langmix

    def save(self, *args, **kwargs):
        return self.elan.save(*args, **kwargs)

    @staticmethod
    def read_eaf(eaf_path, **kwargs):
        ''' Read an EAF file as a Bela2 object

        :param eaf_path: Path to the EAF file
        :type eaf_path: str-like object or a Path object
        :returns: A Bela2 object
        :rtype: bela.Bela2
        '''
        return Bela2(elan.read_eaf(eaf_path), path=eaf_path, **kwargs)

    @staticmethod
    def from_elan(elan, eaf_path=":memory:", **kwargs):
        ''' Create a BELA-con version 2.x object from a :class:`speach.elan.ELANDoc` object '''
        return Bela2(elan, path=eaf_path, **kwargs)


read_eaf = Bela2.read_eaf
from_elan = Bela2.from_elan



# -*- coding: utf-8 -*-

# This code is a part of BELA package: https://github.com/letuananh/bela
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

"""
BELA version 1.x convention
"""

from collections import defaultdict as dd
from collections import OrderedDict

from chirptext import chio
from chirptext.anhxa import DataObject
from speach.ttlig import IGRow
from speach.elan import parse_eaf_stream
from speach.vtt import sec2ts

from .common import TIER_PARSER
from .common import UTTERANCE_GAP_THRESHOLD
from .common import KNOWN_LANGUAGES
from .common import LanguageMix


class Transcript(DataObject):
    """
    Represent a transcript of a recording, mainly constructed from CSV data)
    """

    def __init__(self):
        self.__sents = []  # utterances sorted by starting time
        self.__tiers = dd(list)

    def insert(self, text, tsfrom, tsto=None, tsduration=None, tier=None, **kwargs):
        ''' Add an annotation chunk to a tier '''
        ig = IGRow(text=text, tsfrom=float(tsfrom), **kwargs)
        if tsto is not None:
            ig.tsto = float(tsto)
            if tsduration is not None:
                expected = round(ig.tsduration, 3)
                if expected != round(float(tsduration), 3):
                    raise ValueError("Inconsistent values for tsfrom ({}), tsto ({}), and tsduration ({}). Expected tsduration=({})".format(tsfrom, tsto, tsduration, expected))
        elif tsduration is not None:
            ig.tsto = ig.tsfrom + float(tsduration)
        if tier is not None:
            ig.tier = tier
        self.__tiers[ig.tier].append(ig)
        self.__sents.append(ig)

    def tier_names(self):
        return tuple(self.__tiers.keys())

    def tier(self, tier_name):
        return self.__tiers[tier_name] if tier_name in self.__tiers else None

    def __len__(self):
        return len(self.__sents)

    def __getitem__(self, idx):
        return self.__sents[idx]

    def sort(self):
        ''' Sort all utterances '''
        self.__sents.sort(key=lambda sent: (sent.tsfrom, sent.tsto))
        for tier in self.__tiers.values():
            tier.sort(key=lambda sent: (sent.tsfrom, sent.tsto))
        return self.__sents

    def tag_language(self, utterance_tier_name, language_tier_name, default_value=''):
        ''' Use text value from language_tier as language to tag utterances
            default_value -- Default language value (defaulted to an empty string)
        '''
        utterance_tier = self.__tiers[utterance_tier_name]
        language_tier = self.__tiers[language_tier_name]
        for u in utterance_tier:
            candidates = []
            for l in language_tier:
                score = u.overlap(l)
                if score > 0:
                    candidates.append((score, l))
                if l.tsfrom > u.tsto:
                    break
            if candidates:
                u.language = max(candidates, key=lambda x: x[0])[1].text.strip()  # .replace(':', '__')
            elif not u.language:
                u.language = default_value
        return utterance_tier

    def join_utterances(self, tier_name=None, tier_class=None):
        ''' Group adjecent utterances together. Return a list of joined utterance lists '''
        _timeline = self.__sents if tier_name is None else self.__tiers[tier_name]
        _utterances = []
        _current = []  # current group
        for idx, s in enumerate(_timeline):
            if tier_class and s.tier_class != tier_class:
                continue
            if idx == 0:
                _current.append(s)
            elif s.speaker == _timeline[idx - 1].speaker and abs(s.tsfrom - _timeline[idx - 1].tsto) < UTTERANCE_GAP_THRESHOLD:
                _current.append(s)
            else:
                # flush
                if _current:
                    _utterances.append(_current)
                _current = [s]
        if _current:
            _utterances.append(_current)
        return _utterances

    @staticmethod
    def from_rows(rows):
        ''' create a Transcript object from CSV rows (list of list) '''
        transcript = Transcript()
        for row in rows:
            if len(row) == 5:
                tier, start_sec, end_sec, dur_sec, text = row
                transcript.insert(text, start_sec, tsto=end_sec, tsduration=dur_sec, tier=tier)
            elif len(row) == 6:
                tier, speaker, start_sec, end_sec, dur_sec, text = row
                transcript.insert(text, start_sec, tsto=end_sec, tsduration=dur_sec, tier=tier, speaker=speaker)
            else:
                getLogger().warning(f"Invalid line {row}")
                continue
        return transcript

    @staticmethod
    def read_tsv(file_path, *args, **kwargs):
        return Transcript.from_rows(chio.read_tsv_iter(file_path, *args, **kwargs))


class Bela1:
    ''' This class represent BELA convention version 1'''

    def __init__(self):
        """
        All information about an ELAN transcript
        """
        self.persons = set()  # all persons in this transcript
        self.person_code_map = dd(set)  # a list of all associated codes of this person
        self.person_duration = dd(float)
        self.person_languages = dd(lambda: dd(float))
        self.person_utterances = dd(list)
        self.language_duration = dd(float)
        self.person_warnings = dd(set)
        self.tiers = set()    # all tiers in this transcript
        self.person_tiers = dd(set)
        self.weird_names = set()  # detect weird tier names

        self.tokens = set()
        self.words = set()
        self.languages = set()
        self.others = set()
        self.sents = []
        self.csw = None  # all CSV rows as IGRow
        self.filepath = None

    def sorted_person_tiers(self):
        d = dict()
        for k in sorted(self.person_tiers.keys()):
            d[k] = sorted(self.person_tiers[k])
        return d

    def code_person_map(self):
        cp_map = OrderedDict()
        for p, codes in self.person_code_map.items():
            for c in codes:
                if c in cp_map:
                    raise Exception(f"Duplicated code {c} for person {p} -- All: {codes}")
                else:
                    cp_map[c] = p
        return cp_map

    def sort_languages(self, languages=None):
        if languages is None:
            languages = self.languages
        return sorted(languages, key=lambda i: (i not in KNOWN_LANGUAGES, i != 'Red Dot', i))

    def to_dict(self):
        people = []
        for person in sorted(self.persons):
            person_languages = [(lang, round(duration, 2)) for lang, duration in self.person_languages[person].items()]
            person_languages.sort(key=lambda x: x[1])
            person_languages.reverse()
            person_dict = {'name': person,
                           'code': list(self.person_code_map[person]),
                           'duration': round(self.person_duration[person], 2),
                           'languages': person_languages,
                           'tiers': list(self.person_tiers[person]),
                           'utterance_count': len(self.person_utterances[person])}
            people.append(person_dict)
        return {
            'languages': [(l, round(self.language_duration[l], 2)) for l in self.sort_languages()],
            'people': people
        }

    def to_language_mix(self, to_ts=None, auto_compute=True):
        ''' Collapse utterances to generate a language mix timeline '''
        langmix = LanguageMix()
        for _person_name, _utterances in self.person_utterances.items():
            for u in _utterances:
                if u.chunks is not None:
                    for c in u.chunks:
                        if c.tsfrom and c.tsto:
                            if to_ts is not None and c.tsto > to_ts:
                                continue
                            else:
                                # getLogger().debug(f"{c.ID} -- {c.value} [{c.from_ts.value} - {c.to_ts.value}] | {c.language}")
                                c.duration = c.tsduration
                                langmix.add(c)
                elif u.tsfrom and u.tsto:
                    u.duration = u.tsduration
                    langmix.add(u)
        return langmix.compute() if auto_compute else langmix

    @staticmethod
    def read(filepath, autotag=True):
        ''' Read ELAN csv file '''
        _transcript = Bela1()
        _transcript.filepath = filepath
        _transcript.csw = Transcript.read_tsv(filepath)
        return Bela1.process_transcript(_transcript, autotag=autotag)

    @staticmethod
    def read_eaf(eaf_path):
        elan = parse_eaf_stream(eaf_path)
        return Bela1.from_elan(elan, eaf_path)

    @staticmethod
    def from_elan(elan, eaf_path=":memory:"):
        elanplus = Bela1.parse_rows(elan.to_csv_rows(), filepath=eaf_path)
        elanplus.elan = elan  # store pointer to ELAN object
        return elanplus

    @staticmethod
    def parse_rows(rows, autotag=True, filepath=':memory:'):
        _transcript = Bela1()
        _transcript.filepath = filepath
        _transcript.csw = Transcript.from_rows(rows)
        return Bela1.process_transcript(_transcript, autotag=autotag)

    @staticmethod
    def process_transcript(_transcript, autotag=True):
        for name in _transcript.csw.tier_names():
            m = TIER_PARSER.match(name)
            if m:
                person = m.group('person')
                tier = m.group('tier')
                _transcript.persons.add(person)
                _transcript.tiers.add(tier)
            elif name in ("ActivityMarkers"):
                pass
            else:
                _transcript.weird_names.add(name)
        # auto tag languages
        if autotag:
            tier_names = _transcript.csw.tier_names()
            for person in _transcript.persons:
                utterance_tier = "{} (Utterance)".format(person)
                language_tier = "{} (Language)".format(person)
                if utterance_tier in tier_names and language_tier in tier_names:
                    _transcript.csw.tag_language(utterance_tier, language_tier, "#!#?")
        # extract corpus
        for sent in _transcript.csw:
            m = TIER_PARSER.match(sent.tier)
            _person_name = None
            tsduration = sent.tsduration
            if m:
                _person_name = m.group('person')
                sent.speaker_name = _person_name
                _transcript.person_tiers[_person_name].add(m.group('tier'))
                sent.tier_class = m.group('tier')
                if sent.speaker:
                    _transcript.person_code_map[m.group('person')].add(sent.speaker)
            if sent.tier == 'Transcriber (Comment)':
                # ignore all Transcriber comment
                continue
            if sent.tsduration < 0:
                _transcript.person_warnings[_person_name].add('Negative utterances ({:.2f})'.format(sent.tsduration))
                tsduration = 0
            if sent.text.strip() == '':
                warn_text = "Blank: {} [{} -- {}]".format(_person_name, sec2ts(sent.tsfrom), sec2ts(sent.tsto))
                _transcript.person_warnings[_person_name].add(warn_text)
            if sent.tier_class in ('Utterance', 'Comment'):
                if _person_name and sent.tsduration:
                    _transcript.person_duration[_person_name] += tsduration
                    _transcript.person_utterances[_person_name].append(sent)
                if sent.language in ('', '#!#?'):
                    warn_text = "language???: {} [{} -- {}] {}".format(_person_name, sec2ts(sent.tsfrom), sec2ts(sent.tsto), sent.text)
                    _transcript.person_warnings[_person_name].add(warn_text)

                txt = sent.text.strip()
                _transcript.sents.append(txt)
                _tks = [x.strip() for x in txt.split()]
                _transcript.tokens.update(_tks)
                _transcript.words.update(_tks)
            elif sent.tier_class == 'Language':
                lng = sent.text.strip()
                _transcript.tokens.update((lng,))
                _transcript.languages.update((lng,))
                _transcript.language_duration[lng] += tsduration
                if _person_name:
                    _transcript.person_languages[_person_name][lng] += tsduration
            else:
                _transcript.tokens.update(_tks)
                _transcript.others.update(_tks)
        return _transcript


def build_utterances_json(elanplus):
    ''' Extract all utterances from an Bela1 object, join them by tier_class "Utterance"
    and then return a JSON-ready dictionary '''
    sents = []
    for us in elanplus.csw.join_utterances(tier_class="Utterance"):
        utterances = []
        for u in us:
            utterances.append({k: v for k, v in u.to_dict().items() if v != ""})
        sents.append(utterances)
    return sents



# -*- coding: utf-8 -*-

# This code is a part of BELA package: https://github.com/letuananh/bela
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

"""
Lexical Analyser
"""

import logging
from pathlib import Path
from collections import defaultdict as dd
from collections import Counter

from chirptext import chio
from chirptext import ttl

from .common import tokenize
from .common import NLTK_AVAILABLE
from .common import _process_token, InvalidTokenException


def read_lexicon(lex_name):
    p = Path(__file__).parent / 'data' / lex_name
    forms = set()
    with chio.open(p) as _lex_stream:
        for line in _lex_stream:
            if line.startswith("#"):
                continue
            else:
                forms.add(line.strip())
    return forms


# TODO add all BELA special keywords
_KEYWORDS = ['babyname', 'adultname', 'siblingname', 'strangername',
             '...', 'english', 'chinese', 'malay', 'tamil']
if NLTK_AVAILABLE:
    from nltk.corpus import words  # English words
    from nltk.corpus import stopwords
    _ENGLISH_WORDS = set(words.words())
    _ENGLISH_WORDS.update(stopwords.words('english'))
else:
    _ENGLISH_WORDS = set()
_ENGLISH_WORDS.update(("'s", "ok", "'m", "'ll", "n't", "okay", "'re", "'d", "'ve"))
_MANDARIN_WORDS = read_lexicon('cmn_lex.txt.gz')
_MALAY_WORDS = read_lexicon('msa_lex.txt.gz')


# [2021-03-11 木 11:48]
# Adopted from https://github.com/letuananh/lelesk/blob/master/lelesk/util.py
# LeLESK: MIT License
if NLTK_AVAILABLE:
    # from nltk.tokenize import word_tokenize
    from nltk import pos_tag
    from nltk.stem import WordNetLemmatizer

    wnl = WordNetLemmatizer()


def ptpos_to_wn(ptpos, default='x'):
    ''' Penn Treebank Project POS to WN '''
    if ptpos.startswith('JJ'):
        return 'a'
    elif ptpos.startswith('NN'):
        return 'n'
    elif ptpos.startswith('RB'):
        return 'r'
    elif ptpos.startswith('VB'):
        return 'v'
    else:
        return default


def _tokenize(words):
    if NLTK_AVAILABLE:
        tags = pos_tag(words)
        tokens = [(w, t, wnl.lemmatize(w, pos=ptpos_to_wn(t, default='n'))) for w, t in tags]
        return tokens
    else:
        return words


class LexicalAnalyser:
    def __init__(self, lang_lex_map=None, word_only=False, ellipsis=False, non_word='', lemmatizer=True, **kwargs):
        self.utterances = ttl.Document()
        self.word_sent_map = dd(set)
        self.lang_word_sent_map = dd(lambda: dd(set))
        self.lang_word_speaker_map = dd(lambda: dd(set))
        self.word_only = word_only
        self.ellipsis = ellipsis
        self.non_word = non_word
        self.lemmatizer = lemmatizer
        # setup built-in language-lexicon map
        self.lang_lex_map = {
            'English': _ENGLISH_WORDS,
            'Mandarin': _MANDARIN_WORDS,
            'Malay': _MALAY_WORDS
        }
        # allow custom language_map
        self.__custom_lang_lex_map = lang_lex_map if lang_lex_map else {}
        self.word_speaker_map = dd(set)
        self.word_map = dd(Counter)

    def analyse(self, external_tokenizer=True):
        self.word_sent_map.clear()
        self.word_map.clear()
        for utterance in self.utterances:
            language = utterance.tag.language.value
            speaker = utterance.tag.speaker.value
            # source = utterance.tag.source.value
            tokens = [t.lower() for t in tokenize(
                utterance.text, language=language,
                ellipsis=self.ellipsis, non_word=self.non_word,
                word_only=self.word_only, nlp_tokenizer=external_tokenizer)]
            self.word_map[language].update(tokens)
            for token in tokens:
                self.word_speaker_map[token] = speaker
                self.word_sent_map[token].add(utterance.text)
                self.lang_word_speaker_map[language][token].add(speaker)
                self.lang_word_sent_map[language][token].add(utterance.text)

    def gen_type_token_map(self):
        ratio_map = {}
        for lang, counter in self.word_map.items():
            count_token = len(list(counter.elements()))
            count_type = len(counter)
            ratio = count_token / count_type if count_type > 0 else 0
            ratio_map[lang] = (count_token, count_type, ratio)
        return ratio_map

    def gen_type_token_list(self):
        _list = [(lang, count_token, count_type, ratio) for lang, (count_token, count_type, ratio) in self.gen_type_token_map().items()]
        _list.sort(key=lambda x: -x[3])
        return _list

    def add(self, text, language, **kwargs):
        sent = self.utterances.sents.new(text)
        sent.tag.language = language
        for k, v in kwargs.items():
            sent.tag[k] = v

    def is_special_token(self, word, language):
        ''' Determine if a given token is a special token (keywords, markup, etc.) '''
        return word == '###' or word.startswith(':')

    def is_unknown(self, word, language):
        '''Check if a word is a known word (exists in the current lexicon)'''
        if word in _KEYWORDS:
            return False
        elif language in self.lang_lex_map:
            if word in self.lang_lex_map[language]:
                return False
            elif language not in self.__custom_lang_lex_map:
                return True
            else:
                return word not in self.__custom_lang_lex_map[language]
        elif language in self.__custom_lang_lex_map:
            return word not in self.__custom_lang_lex_map[language]
        else:
            return False

    def to_dict(self, ignore_empty=True):
        stats_dict = {'languages': [], 'lexicon': [], 'errors': []}
        __lemmatize_error = False
        for lang, count_token, count_type, ratio in self.gen_type_token_list():
            if ignore_empty and not count_type and not count_token and not ratio:
                continue
            stats_dict['languages'].append({
                'language': lang,
                'types': count_type,
                'tokens': count_token,
                'ratio': round(ratio, 2)
            })
        for lang, counter in self.word_map.items():
            lang_lexicon = {'language': lang, 'vocabs': []}
            for word, freq in counter.most_common():
                _is_special = self.is_special_token(word, lang)
                if _is_special:
                    try:
                        _process_token(word)
                        _is_unknown = False
                    except InvalidTokenException:
                        _is_unknown = True
                else:
                    lemma = word
                    # try to lemmatize if possible
                    if NLTK_AVAILABLE and self.lemmatizer and not __lemmatize_error and lang == 'English':
                        try:
                            __, tag = pos_tag([word])[0]
                            lemma = wnl.lemmatize(word, pos=ptpos_to_wn(tag, default='n'))
                        except Exception as e:
                            # logging.getLogger(__name__).exception("BELA.Lemmatizer crashed")
                            # do not lemmatize if NLTK crashed
                            __lemmatize_error = True
                            if isinstance(e, LookupError):
                                if 'omw-1.4' in str(e):
                                    stats_dict['errors'].append(f'Lexicon was generated without lemmatizer. OMW-1.4 data not found.')
                                else:
                                    stats_dict['errors'].append(f'Lexicon was generated without lemmatizer. Unknown resource missing.')
                            else:
                                stats_dict['errors'].append('Lexicon was generated without lemmatizer. Unknown error was raised.')
                    _is_unknown = self.is_unknown(lemma, lang)
                _lex_entry = {
                    'word': word,
                    'freq': freq,
                    'sents': list(self.lang_word_sent_map[lang][word]),
                    'speakers': list(self.lang_word_speaker_map[lang][word]),
                    'special_code': _is_special,
                    'unknown_word': _is_unknown
                }
                lang_lexicon['vocabs'].append(_lex_entry)
            if not ignore_empty or lang_lexicon['vocabs']:
                stats_dict['lexicon'].append(lang_lexicon)
        return stats_dict


class CorpusLexicalAnalyser:
    ''' Analyse a corpus text '''
    def __init__(self, filepath=':memory:', lang_lex_map=None, word_only=False, lemmatizer=True, **kwargs):
        self.filepath = filepath
        self.word_only = word_only
        self.lemmatizer = lemmatizer
        self.__lang_lex_map = {} if lang_lex_map is None else lang_lex_map
        self.profiles = dd(self._create_lex_analyzer)

    def _create_lex_analyzer(self):
        return LexicalAnalyser(lang_lex_map=self.__lang_lex_map,
                               word_only=self.word_only,
                               lemmatizer=self.lemmatizer)

    def read(self, **kwargs):
        ''' Read the CSV file content specified by self.filepath '''
        for text, language, source, speaker in chio.read_csv_iter(self.filepath, **kwargs):
            self.add(text, language, source=source, speaker=speaker)
        return self

    def add(self, text, language, source='', speaker=''):
        if text is None:
            text = ''
        if language is None:
            language = ''
        if source is None:
            source = ''
        if speaker is None:
            speaker = ''
        self.profiles['ALL'].add(text, language, source=source, speaker=speaker)
        self.profiles[speaker].add(text, language, source=source, speaker=speaker)

    def analyse(self, external_tokenizer=True):
        ''' Analyse all available profiles (i.e. speakers) '''
        for profile in self.profiles.values():
            profile.analyse(external_tokenizer=external_tokenizer)
        return self

    def to_dict(self):
        ''' Export analysed result as a JSON-ready object '''
        profile_list = []
        for pname in sorted(self.profiles.keys()):
            profile = self.profiles[pname]
            profile_list.append({
                'name': pname,
                'stats': profile.to_dict()
            })
        return profile_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Setup script for BELA.

Latest version can be found at https://github.com/letuananh/bela

:copyright: (c) 2019 Le Tuan Anh, BLIP lab NTU, Singapore <tuananh.ke@gmail.com>
:license: MIT, see LICENSE for more details.
'''

import io
from setuptools import setup


def read(*filenames, **kwargs):
    ''' Read contents of multiple files and join them together '''
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


readme_file = 'README.md'
long_description = read(readme_file)
pkg_info = {}
exec(read('bela/__version__.py'), pkg_info)

with open('requirements.txt', 'r') as infile:
    requirements = infile.read().splitlines()

setup(
    name='bela',  # package file name (<package-name>-version.tar.gz)
    version=pkg_info['__version__'],
    url=pkg_info['__url__'],
    project_urls={
        "Bug Tracker": "https://github.com/letuananh/bela/issues",
        "Source Code": "https://github.com/letuananh/bela/"
    },
    keywords=["corpus", "linguistics", "multilingual", "transcription",
              "ELAN", "NLP", "children", "language acquisition",
              "parent-child", "conversation", "discourse analysis"],
    license=pkg_info['__license__'],
    author=pkg_info['__author__'],
    tests_require=[],
    install_requires=requirements,
    python_requires=">=3.5",
    author_email=pkg_info['__email__'],
    description=pkg_info['__description__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['bela', 'bela.data'],
    package_data={'bela': ['data/*.gz', 'data/*.eaf']},
    include_package_data=True,
    platforms='any',
    test_suite='test',
    # Reference: https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=['Programming Language :: Python',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: 3.8',
                 'Programming Language :: Python :: 3.9',
                 'Programming Language :: Python :: 3.10',
                 'Programming Language :: Python :: 3.11',
                 'Development Status :: {}'.format(pkg_info['__status__']),
                 'Natural Language :: English',
                 'Natural Language :: Chinese (Simplified)',
                 'Natural Language :: Malay',
                 'Natural Language :: Tamil',
                 'Environment :: Plugins',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: {}'.format(pkg_info['__license__']),
                 'Operating System :: OS Independent',
                 'Topic :: Text Processing',
                 'Topic :: Software Development :: Libraries :: Python Modules']
)



# -*- coding: utf-8 -*-

# This code is a part of BELA package: https://github.com/letuananh/bela
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

"""
Common functions
"""

import re
import logging
import string

try:
    from nltk import word_tokenize
    NLTK_AVAILABLE = True
except ModuleNotFoundError:
    NLTK_AVAILABLE = False
try:
    import jieba
    JIEBA_AVAILABLE = True
except ModuleNotFoundError:
    JIEBA_AVAILABLE = False
from speach.vtt import sec2ts


class InvalidTokenException(ValueError):
    ''' InvalidTokenException is raised when a text token is not a valid BELA token '''
    pass


# ... for ellipsis
PUNCS = list(string.punctuation) + list(' ，。？（）！') + ["؟"] + ['``', "''", '...']


def _remove_punc(tokens):
    return (x for x in tokens if x not in PUNCS)


def getLogger():
    return logging.getLogger(__name__)


def maketime(s):
    return f"{sec2ts(s.tsfrom)} - {sec2ts(s.tsto)} ({s.tsduration:02})"


def getlang(x):
    return x.strip() if x != 'N/A' else ''


def is_special_token(t):
    return "<" in t or ":" in t or "/" in t or "#" in t or "=" in t


KNOWN_LANGUAGES = ['English', 'Mandarin', 'Tamil', 'Cantonese', 'Malay', 'Javanese', 'Arabic', 'Hokkien', 'Red Dot']
# [2020-09-30] To take Interjection out of KNOWN_LANGUAGE_CLASSES to assign a label
KNOWN_LANGUAGE_CLASSES = ['Vocal Sounds', 'Non Vocal Sounds', 'Inaudible', 'Languageless', 'Interjection'] + [':v:airstream', ':v:crying', ':v:vocalizations', ':v:laughter']
UTTERANCE_GAP_THRESHOLD = 0.05
DEFAULT_NON_WORD = 'XbeepX'
TIER_PARSER = re.compile(r'(?P<person>[\w ]+) +\((?P<tier>[\w ]+)\)')
# usage
# m = TIER_PARSER.match('baby (Comments)')
BELA_CLASS_LABEL = re.compile("^[a-z0-9_]+$")
PTN_ELLIPSIS = re.compile(r'(\.{2,})$')
PTN_INNER_STOP = re.compile(r'\.+[^$\.?]+$')
VOCAL_CLOSED_CLASS = {
    'crying', 'kissing_sound', 'tsk', 'airstream', 'coughing',
    'animal_sound', 'flamingo_sound', 'orangutan_sound', 'elephant_sound',
    'lightning_sound', 'splash_sound', 'thunder_sound', 'rain_sound',
    'ringing_sound', 'screaming_sound', 'shushing_sound',
    'train_sound', 'gasp', 'laughter',
    'sigh', 'vocalizations', 'whistle', 'x'}
#  [2021-08-30 月 16:54] add 6 new vocal sounds
# 'flamingo_sound', 'orangutan_sound',
# 'lightning_sound ', 'splash_sound',
# 'ringing_sound', 'screaming_sound'
VOCAL_CLOSED_CLASS_3 = {'r', 'u'}
NON_VOCAL_SOUNDS = {'clapping', 'sneeze', 'clicking', 'sniff', 'sync', 'x'}


def _process_token(token, non_word=DEFAULT_NON_WORD, mimicked=False, word_only=True, **kwargs):
    ''' Extract text from a special ELAN token '''
    _res = [':m'] if mimicked and not word_only else []
    if token.startswith(":"):
        parts = [x for x in token.strip().split(":") if x]
        if not parts:
            raise InvalidTokenException("Token text is blank")
        elif parts[0] == 'v':
            if len(parts) == 2:
                if parts[1] in VOCAL_CLOSED_CLASS:
                    if word_only:
                        _res.append(non_word)
                    else:
                        _res.append(token.strip())
                else:
                    raise InvalidTokenException("Invalid vocal sounds tag")
            elif len(parts) == 3:
                if parts[1] in VOCAL_CLOSED_CLASS_3:
                    if not word_only:
                        _res.append(f":{parts[0]}:{parts[1]}:")
                    _res.extend(parts[2].split('+'))
                elif parts[1] == 'l':
                    # [2021-08-30 月 16:58] add labelled vocal sounds
                    if BELA_CLASS_LABEL.match(parts[2]):
                        if word_only:
                            _res.append(non_word)
                        else:
                            _res.append(token.strip())
                    else:
                        raise InvalidTokenException("This lablled vocal sound token contains invalid character(s)")
                else:
                    raise InvalidTokenException("invalid unclassifiable vocal sound token")
        elif parts[0] == 'm':
            if mimicked:
                raise InvalidTokenException("Recursive mimicking is not allowed")
            if len(parts) == 2:
                if not word_only:
                    _res.append(":m")
                _res.extend(parts[1].split('+'))  # Mimic(text)
            else:
                _rest = token[token.index(":m")+2:]
                return _process_token(_rest, mimicked=True, non_word=non_word, word_only=word_only, **kwargs)  # mimicked structure
        elif len(parts) == 2:
            if parts[0] == 's':
                if parts[1] in NON_VOCAL_SOUNDS:
                    if word_only:
                        _res.append(non_word,)
                    else:
                        _res.append(f":{parts[0]}:{parts[1]}:")
                else:
                    raise InvalidTokenException("Invalid non vocal sound, closed class token")
            if parts[0] in ('si', "m"):
                if not word_only:
                    _res.append(":" + parts[0])
                _res.extend(parts[1].split("+"))
            if parts[0] == 'd':
                if not word_only:
                    _res.append(":d")
                _res.append(parts[1])
        elif len(parts) == 3:
            if parts[0] == 'l':
                if not word_only:
                    _res.append(":l:" + parts[1])
                _res.extend(parts[2].split('+'))
    elif token.startswith('='):
        if not word_only:
            _res.append("=")
        _res.append(token[1:].strip())
    elif token in ('##', '###'):
        if word_only:
            _res.append(non_word)
        else:
            _res.append(token)
    elif "=" in token:
        _res.extend(token.split("="))
    return _res


def _report_error(message, errors=None):
    if errors is not None:
        errors.append(message)
    else:
        getLogger().warning(message)


INVALID_CHAR = r'\{\}\[\]\r\n\t,ɡʔʲ‘’'
FULLWIDTH_CHARS = r'，。？（）！　'
# [2022-01-06 木 14:05] TA
# Allow ? in Mandarin annotations for now
HALFWIDTH_CHARS = r'\(\)'  # ?
OTHER_LANG_INVALID = re.compile(fr'[{INVALID_CHAR}]')
ENGLISH_INVALID = re.compile(fr'[{INVALID_CHAR}{FULLWIDTH_CHARS}]')
MANDARIN_INVALID = re.compile(fr'[{INVALID_CHAR}{HALFWIDTH_CHARS}]')


def _contain_invalid_characters(text, language=None):
    if not language:
        return OTHER_LANG_INVALID.search(text)
    elif language == 'Mandarin':
        return MANDARIN_INVALID.search(text)
    else:
        return ENGLISH_INVALID.search(text)


def _find_invalid_characters(text, language=None):
    if not language:
        return OTHER_LANG_INVALID.findall(text)
    elif language == 'Mandarin':
        return MANDARIN_INVALID.findall(text)
    else:
        return ENGLISH_INVALID.findall(text)


class TokenList:
    def __init__(self, *args, **kwargs):
        self.__tokens = list(*args, **kwargs)

    def __eq__(self, other):
        return self.__tokens == other

    def __len__(self):
        return len(self.__tokens)

    def __iter__(self):
        return iter(self.__tokens)

    def append(self, token):
        if token is not None:
            # strip off special characters ("=", "~", etc.)
            if token.startswith('='):
                token = token[1:]
            if token.endswith('~'):
                token = token[:-1]
            if token:
                self.__tokens.append(token)

    def extend(self, tokens):
        for t in tokens:
            self.append(t)

    def _append_nlp(self, token, language, nlp_tokenizer=True):
        if nlp_tokenizer:
            self.extend(_nlp_tokenize(token, language=language))
        else:
            self.append(token)

    def _extend_nlp(self, tokens, language, nlp_tokenizer=True):
        for token in tokens:
            self._append_nlp(token, language, nlp_tokenizer=nlp_tokenizer)

    def __repr__(self):
        return f"TokenList({repr(self.__tokens)})"

    def __str__(self):
        return str(self.__tokens)

    def to_list(self):
        return list(self.__tokens)

    def to_tuple(self):
        return tuple(self.__tokens)

    def remove_punc(self):
        self.__tokens = list(_remove_punc(self.__tokens))
        return self


PTN_PRETOKENIZER = re.compile(r'([^ ?　？؟]+|[\?？؟])')


def __pretokenize(text):
    return PTN_PRETOKENIZER.findall(text)


def _nlp_tokenize(text, language='English'):
    if language == 'English' and NLTK_AVAILABLE:
        words = word_tokenize(text)
    elif language == 'Mandarin' and JIEBA_AVAILABLE:
        words = jieba.cut(text, cut_all=False)
    else:
        words = __pretokenize(text)
    return words


def _bela_tokenize(text, non_word=DEFAULT_NON_WORD, errors=None, ellipsis=True,
                   language='English', nlp_tokenizer=True,
                   remove_punc=True, **kwargs):
    ''' Convert MMC convention to normal token text
        >>> process_text(':si:quack+quack+quack+quack+quack tada! :s:clapping finish!')
        >>> 'quack quack quack quack quack tada! XbeepX finish!'
        errors: must be a list, or an object that supports append() function
    '''
    tokens = __pretokenize(text)
    new_tokens = TokenList()
    for token in tokens:
        if is_special_token(token):
            try:
                pieces = _process_token(token, non_word=non_word, **kwargs)
            except InvalidTokenException as e:
                emsg = getattr(e, 'message', str(e))
                _report_error(f"{emsg} ({token})", errors=errors)
                continue
            # except Exception:
            #    pieces = []
            if pieces:
                if ellipsis:
                    _selected = (x for x in pieces if x)
                else:
                    _selected = tuple(x for x in pieces if x and not x.endswith('...'))  # remove ellipsis
                for p in _selected:
                    if p[0] == ':':
                        # do not split special token any further
                        new_tokens.append(p)
                    else:
                        new_tokens._append_nlp(p, language=language, nlp_tokenizer=nlp_tokenizer)
            else:
                _report_error("Invalid token ({})".format(token), errors=errors)
        elif token.endswith('~'):
            new_tokens._append_nlp(token[:-1], language=language, nlp_tokenizer=nlp_tokenizer)
        elif token.startswith("~") or "~" in token:
            _report_error(f"Tildes can only be placed at the end of a token ({token})", errors=errors)
            new_tokens._append_nlp(token, language=language, nlp_tokenizer=nlp_tokenizer)
        elif token.strip() == '...':
            _report_error(f"Ellipsis markers (...) must not follow empty space", errors=errors)
        elif PTN_ELLIPSIS.match(token.strip()):
            _report_error(f"Unknown dots ({token})", errors=errors)
        else:
            # validate ellipsis
            _mi = PTN_INNER_STOP.search(token)
            if _mi:
                _report_error(f"Invalid punctuation ({token})", errors=errors)
            _m = PTN_ELLIPSIS.search(token)
            if _m:
                if len(_m.group(1)) != 3:
                    _report_error(f"Ellipses are denoted by exact 3 full stops ({token} was found)", errors=errors)
                    new_tokens._append_nlp(token, language=language, nlp_tokenizer=nlp_tokenizer)
                elif ellipsis:
                    # this is a valid ellipsis token
                    new_tokens._append_nlp(token[:-3], language=language, nlp_tokenizer=nlp_tokenizer)
            else:
                # normal token
                new_tokens._append_nlp(token, language=language, nlp_tokenizer=nlp_tokenizer)
    return new_tokens.remove_punc() if remove_punc else new_tokens


def process_text(text, non_word=DEFAULT_NON_WORD, errors=None, ellipsis=True, **kwargs):
    ''' Convert MMC convention to normal token text
        >>> process_text(':si:quack+quack+quack+quack+quack tada! :s:clapping finish!')
        >>> 'quack quack quack quack quack tada! XbeepX finish!'
        errors: must be a list, or an object that supports append() function
    '''
    new_tokens = _bela_tokenize(text, non_word=DEFAULT_NON_WORD, errors=errors, ellipsis=ellipsis, **kwargs)
    return " ".join(t.strip() for t in new_tokens)


tokenize = _bela_tokenize


class LanguageMix:
    def __init__(self):
        self.__parts = []  # actual utterance chunks
        self.chunks = []  # collapsed chunks (for language-mix diagram)
        self.length = 0

    def add(self, c):
        ''' Store an utterance chunk '''
        self.__parts.append(c)

    def sort(self):
        self.__parts.sort(key=lambda x: (x.from_ts, x.to_ts))

    def compute(self, auto_sort=True):
        if auto_sort:
            self.sort()
        self.chunks.clear()
        for c in self.__parts:
            # getLogger().debug(f"{c.ID} -- {c.value} [{c.from_ts.value} - {c.to_ts.value}] | {c.language}")
            if self.chunks and self.chunks[-1][0] == c.language:
                self.chunks[-1][1] += c.duration
            else:
                self.chunks.append([c.language, c.duration])
        self.length = sum(x.duration for x in self.__parts)
        return self

    def to_dict(self):
        return {'chunks': self.chunks, 'length': self.length}
